#include <stdarg.h>

#include "doomtype.h"
#include "doomstat.h"
#include "d_player.h"
#include "s_sound.h"
#include "gi.h"
#include "d_net.h"
#include "g_game.h"
#include "p_local.h"
#include "sv_main.h"
#include "i_system.h"
#include "c_console.h"
#include "c_dispatch.h"
#include "m_argv.h"
#include "m_random.h"
#include "s_sndseq.h"
#include "v_video.h"

int     lastreliablenum;
char   parse_cl;

client_t clients[MAXPLAYERS + 1];

int       maxclients;
client_t  tempclient;

#define CLIENT_TIMEOUT 65 // 65 seconds

// Different welcome strings
char welcomestring[] = "Version 1.4a Server\n";
char motdstring[] = "Current csDoom client: www.voxelsoft.com/csdoom\n";
char guidstring[] = "GUID{A7EDB62B-4CDE-4c89-8383-B8B26E6B4F63}\n";
char welcomestring_teamplay[] = "Version 1.4a Server - TEAMPLAY\n";

int  gametime;

void SV_UpdateConsolePlayer(client_t* cl);
void SV_Fraglogfile (void);
int SV_ChooseTeamForPlayer(player_t *player);
void P_SpawnPlayer (mapthing2_t* mthing);

CVAR (rcon_password, "", CVAR_SERVERINFO)

BEGIN_COMMAND (fraglog)
{
	SV_Fraglogfile();
}
END_COMMAND (fraglog)

//
// SV_InitNetwork
//
void SV_InitNetwork (void)
{
    char      *v;

	netgame = false;  // for old network code
	multiplayer = true;

 
	v = Args.CheckValue ("-port");
    if (v)
    {
       localport = atoi (v);
       Printf (PRINT_HIGH, "using alternate port %i\n", localport);
    }
	else 
	   localport = SERVERPORT;
	
	// set up a socket and net_message buffer
	InitNetCommon();

	// determine my name & address
	// NET_GetLocalAddress ();

	Printf(PRINT_HIGH, "UDP Initialized\n");

	v = Args.CheckValue ("-maxclients");
    if (v)
    {
		maxclients = atoi(v);
		if (maxclients < 0)
			maxclients = 0;
		if (maxclients > MAXPLAYERS)
			maxclients = MAXPLAYERS;
	}
	else
		maxclients = DEFPLAYERS;

	v = Args.CheckValue("-master");
	if(v)
	{
		SV_AddMaster(v);
	}

	// initialize clients
	for (int i=0; i<MAXPLAYERS; i++)
    {
		clients[i].netbuf.data = new byte[MAX_UDP_PACKET];
		clients[i].netbuf.maxsize = MAX_UDP_PACKET;
		SZ_Clear(&clients[i].netbuf);
	   
		clients[i].reliablebuf.data = new byte[MAX_UDP_PACKET];
		clients[i].reliablebuf.maxsize = MAX_UDP_PACKET;
		SZ_Clear(&clients[i].reliablebuf);

		clients[i].relpackets.data = new byte[1024*50];
		clients[i].relpackets.maxsize = 1024*50; // 50 KB

		clients[i].cls = cs_free;
	}

	tempclient.reliablebuf.data = new byte[MAX_UDP_PACKET];
    tempclient.reliablebuf.maxsize = MAX_UDP_PACKET;

	gametime = I_GetTime ();
	
	SV_InitMaster();	
}


int SV_GetFreeClient(void)
{
	int i;

	for (i=0; i < maxclients; i++)
	{
		if (clients[i].cls == cs_free)
			return i;
	}
	
	return -1;
}

short SV_FindClientByAddr(void)
{
    int i;

	// DL - Make sure this is is not a blank ip [todo: also limit number of ppl from a single ip to prevent crackers]
	if(net_from.ip[0] == 0 && net_from.ip[1] == 0 && net_from.ip[2] == 0 && net_from.ip[3] == 0)
		return -1;

	for (i=0; i< maxclients; i++)
	{
	   if (clients[i].cls == cs_free)
	       continue;
		
	   if (NET_CompareAdr(clients[i].address, net_from) )
	       return i;
	}
	
	return -1;
}

//
// SV_CheckTimeouts
// If a packet has not been received from a client in CLIENT_TIMEOUT
// seconds, drop the conneciton.
//
void SV_CheckTimeouts (void)
{
	for (int i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls == cs_free)
			continue;

		if (gametic - cl->last_received == CLIENT_TIMEOUT*35)
		{
			parse_cl = i;
		    SV_DropClient();
		}
	}
parse_cl = -1;// changed - added
}

//
// SV_GetPackets
//
bool locked = false;
void SV_GetPackets (void)
{
    while(NET_GetPacket())
    {
        // packet loss emulation :)
        //if (P_Random(pr_misc) > 150)
		//	continue;

        parse_cl = SV_FindClientByAddr();

        if (parse_cl == -1) // no client with net_from address
		{
    	    // apparently, someone is trying to connect
    	    if (gamestate == GS_LEVEL)
				SV_GetChallenge();
			locked = false;
            continue;
        }

		clients[parse_cl].last_received = gametic;

		if (clients[parse_cl].cls == cs_challenge)
		{
			SV_ConnectClient();
			continue;
		}

		if (clients[parse_cl].cls == cs_connected)
		{
			SV_SpawnClient();
			continue;
		}

		SV_ParseCommands();
    }
}

//
// SV_Sound
//
void SV_Sound (AActor *mo, byte channel, char *name, byte attenuation)
{
	int        sfx_id;
	client_t  *cl;
	int        i;

	if (*name == '*')
	{
		// Sexed sound
		char nametemp[128];
		const char templat[] = "player/%s/%s";
		const char *genders[] = { "male", "female", "cyborg" };
		player_t *player;

		sfx_id = -1;
		if ( (player = mo->player) ) 
		{
			sprintf (nametemp, templat, skins[player->userinfo.skin].name, name + 1);
			sfx_id = S_FindSound (nametemp);
			if (sfx_id == -1) 
			{
				sprintf (nametemp, templat, genders[player->userinfo.gender], name + 1);
				sfx_id = S_FindSound (nametemp);
			}
		}
		if (sfx_id == -1) 
		{
			sprintf (nametemp, templat, "male", name + 1);
			sfx_id = S_FindSound (nametemp);
		}
	} 
	else
		sfx_id = S_FindSound (name);


	if (sfx_id > 255 || sfx_id < 0)
	{
		Printf (PRINT_HIGH, "SV_Sound: range error. Sfx_id = %d\n", sfx_id);
		return;
	}

	for (i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		MSG_WriteByte (&cl->netbuf, svc_startsound);
		MSG_WriteShort (&cl->netbuf, mo->netid);
		MSG_WriteByte (&cl->netbuf, channel);
		MSG_WriteByte (&cl->netbuf, sfx_id);
		MSG_WriteByte (&cl->netbuf, attenuation);
	}

}


void SV_Sound (client_t *cl, AActor *mo, byte channel, char *name, byte attenuation)
{
	int sfx_id;

	if (*name == '*')
	{
		// Sexed sound
		char nametemp[128];
		const char templat[] = "player/%s/%s";
		const char *genders[] = { "male", "female", "cyborg" };
		player_t *player;

		sfx_id = -1;
		if ( (player = mo->player) ) 
		{
			sprintf (nametemp, templat, skins[player->userinfo.skin].name, name + 1);
			sfx_id = S_FindSound (nametemp);
			if (sfx_id == -1) 
			{
				sprintf (nametemp, templat, genders[player->userinfo.gender], name + 1);
				sfx_id = S_FindSound (nametemp);
			}
		}
		if (sfx_id == -1) 
		{
			sprintf (nametemp, templat, "male", name + 1);
			sfx_id = S_FindSound (nametemp);
		}
	} 
	else
		sfx_id = S_FindSound (name);

	if (sfx_id > 255 || sfx_id < 0)
	{
		Printf (PRINT_HIGH, "SV_Sound: range error. Sfx_id = %d\n", sfx_id);
		return;
	}

	MSG_WriteByte (&cl->netbuf, svc_startsound);
	if (mo == NULL)
		MSG_WriteShort (&cl->netbuf, 0);
	else
		MSG_WriteShort (&cl->netbuf, mo->netid);
	MSG_WriteByte (&cl->netbuf, channel);
	MSG_WriteByte (&cl->netbuf, sfx_id);
	MSG_WriteByte (&cl->netbuf, attenuation);
}

//
// SV_SoundAvoidCl
// Sends a sound to clients, but doesn't send it to client 'player'.
//
void SV_SoundAvoidCl (player_t *player, AActor *mo, byte channel, char *name, byte attenuation)
{
	int        i;
	int        sfx_id;
	client_t  *cl;

	if (*name == '*')
	{
		// Sexed sound
		char nametemp[128];
		const char templat[] = "player/%s/%s";
		const char *genders[] = { "male", "female", "cyborg" };
		player_t *player;

		sfx_id = -1;
		if ( (player = mo->player) ) 
		{
			sprintf (nametemp, templat, skins[player->userinfo.skin].name, name + 1);
			sfx_id = S_FindSound (nametemp);
			if (sfx_id == -1) 
			{
				sprintf (nametemp, templat, genders[player->userinfo.gender], name + 1);
				sfx_id = S_FindSound (nametemp);
			}
		}
		if (sfx_id == -1) 
		{
			sprintf (nametemp, templat, "male", name + 1);
			sfx_id = S_FindSound (nametemp);
		}
	} 
	else
		sfx_id = S_FindSound (name);

	if (sfx_id > 255 || sfx_id < 0)
	{
		Printf (PRINT_HIGH, "SV_Sound: range error. Sfx_id = %d\n", sfx_id);
		return;
	}

    for (i=0; i<MAXPLAYERS; i++)
    {
		cl = &clients[i];

		if (cl->cls < cs_spawned || player == &players[i])
			continue;

		MSG_WriteByte (&cl->netbuf, svc_startsound);
		MSG_WriteShort (&cl->netbuf, mo->netid);
		MSG_WriteByte (&cl->netbuf, channel);
		MSG_WriteByte (&cl->netbuf, sfx_id);
		MSG_WriteByte (&cl->netbuf, attenuation);
    }
}


void SV_Sound (fixed_t x, fixed_t y, byte channel, char *name, byte attenuation)
{
	int        sfx_id;
	client_t  *cl;
	int        i;

	sfx_id = S_FindSound (name);

	if (sfx_id > 255 || sfx_id < 0)
	{
		Printf (PRINT_HIGH, "SV_Sound: range error. Sfx_id = %d\n", sfx_id);
		return;
	}

	for (i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		MSG_WriteByte (&cl->netbuf, svc_soundorigin);
		MSG_WriteLong (&cl->netbuf, x);
		MSG_WriteLong (&cl->netbuf, y);
		MSG_WriteByte (&cl->netbuf, channel);
		MSG_WriteByte (&cl->netbuf, sfx_id);
		MSG_WriteByte (&cl->netbuf, attenuation);
	}

}


//
// SV_StartSequence
//
void SV_StartSequence (sector_t *sector, int sequence, seqtype_t type)
{
	int sec, i;

	for (sec=0; sec<numsectors; sec++)
		if (sector == &sectors[sec])
			break;

	for (i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		MSG_WriteByte (&cl->reliablebuf, svc_startsectorsequence);
		MSG_WriteShort (&cl->reliablebuf, sec);
		MSG_WriteShort (&cl->reliablebuf, sequence);
		MSG_WriteByte (&cl->reliablebuf, (byte)type);
	}
}

//
// SV_StopSequence
//
void SV_StopSequence(sector_t *sector)
{
	int sec, i;

	for (sec=0; sec<numsectors; sec++)
		if (sector == &sectors[sec])
			break;

	for (i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		MSG_WriteByte (&cl->reliablebuf, svc_stopsectorsequence);
		MSG_WriteShort (&cl->reliablebuf, sec);
	}
}


//
// SV_UpdateFrags
//
void SV_UpdateFrags (player_t *player)
{
    int  playernum;

    for (playernum=0; playernum<MAXPLAYERS; playernum++)
    {
        if (player == &players[playernum])
            break;
    }

    for (int i=0; i<MAXPLAYERS; i++)
    {
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

        MSG_WriteByte (&cl->reliablebuf, svc_updatefrags);
		MSG_WriteByte (&cl->reliablebuf, playernum);
        MSG_WriteShort (&cl->reliablebuf, player->fragcount);
    }
}

//
// SV_SendUserInfo
//
void SV_SendUserInfo (int player, client_t* cl)
{
	player_t *p = &players[player];

	MSG_WriteByte (&cl->reliablebuf, svc_userinfo);
	MSG_WriteByte (&cl->reliablebuf, player);
    MSG_WriteString (&cl->reliablebuf, p->userinfo.netname);
    MSG_WriteString (&cl->reliablebuf, p->userinfo.team);
    MSG_WriteLong (&cl->reliablebuf, p->userinfo.gender);
    MSG_WriteLong (&cl->reliablebuf, p->userinfo.color);
}


bool SV_SetupUserInfo(void)
{
    player_t *p;
	client_t *cl;
	static char     string[2048];
#
    p = &players[parse_cl];
	cl = &clients[parse_cl];

	strcpy(players[parse_cl].userinfo.netname, "noname");

    strcpy(string, MSG_ReadString());// changed
		if(/*!strlen(string) ||*/ strlen(string) > MAXPLAYERNAME || strstr(string, "%"))
			{
			MSG_WriteByte (&cl->reliablebuf, svc_print);
			MSG_WriteByte (&cl->reliablebuf, PRINT_HIGH);
			MSG_WriteString (&cl->reliablebuf, "[server]: invalid nickname\n");
			Printf(PRINT_HIGH, "%s : invalid nickname, client dropped\n", NET_AdrToString(net_from));
			SV_DropClient();
			return false;
			}
	strcpy(p->userinfo.netname, string);

	string[0] = NULL;
    strcpy(string, MSG_ReadString());
		if(strlen(string) > MAXPLAYERNAME || strstr(string, "%"))
			{
			MSG_WriteByte (&cl->reliablebuf, svc_print);
			MSG_WriteByte (&cl->reliablebuf, PRINT_HIGH);
			MSG_WriteString (&cl->reliablebuf, "[server]: invalid team)\n");
			Printf(PRINT_HIGH, "%s : invalid team, client dropped\n", NET_AdrToString(net_from));
			SV_DropClient();
			return false;
			}
	
	if(!teamplay.value || !strlen(p->userinfo.team))
		strcpy(p->userinfo.team, string); // changed

    p->userinfo.gender = MSG_ReadLong();
	if(!teamplay.value)
		p->userinfo.color = MSG_ReadLong();
    p->userinfo.aimdist = MSG_ReadLong();

	return true;
}


//
// SV_ClientFullUpdate
//
void SV_ClientFullUpdate (client_t *cl)
{
	AActor *mo;
	int     i;

	// send players info to the client
	for (i=0; i<MAXPLAYERS; i++)
	{
		if (clients[i].cls < cs_spawned || cl == &clients[i])
			continue;

		MSG_WriteByte(&cl->reliablebuf, svc_spawnplayer);
		MSG_WriteByte(&cl->reliablebuf, i);
		MSG_WriteByte(&cl->reliablebuf, PST_REBORN);
		MSG_WriteShort(&cl->reliablebuf, players[i].mo->netid);
		MSG_WriteLong(&cl->reliablebuf, players[i].mo->angle);
		MSG_WriteLong(&cl->reliablebuf, players[i].mo->x);
		MSG_WriteLong(&cl->reliablebuf, players[i].mo->y);
		MSG_WriteLong(&cl->reliablebuf, players[i].mo->z);

		SV_SendUserInfo(i, cl);
	}

	// update frags
	for (i=0; i<MAXPLAYERS; i++)
	{
		if (!playeringame[i])
			continue;

		if (clients[i].cls < cs_spawned)
			continue;

		MSG_WriteByte(&cl->reliablebuf, svc_updatefrags);
		MSG_WriteByte(&cl->reliablebuf, i);
		MSG_WriteShort(&cl->reliablebuf, players[i].fragcount);
	}

	// update things
	TThinkerIterator<AActor> iterator;
    while ( (mo = iterator.Next() ) )
    {
		if (!mo->netid)
			continue;

        if (mo->type != MT_PLAYER 
			|| (mo->type == MT_PLAYER && !mo->player) )
		{
			MSG_WriteByte(&cl->reliablebuf, svc_spawnmobj);
			MSG_WriteLong(&cl->reliablebuf, mo->x);
			MSG_WriteLong(&cl->reliablebuf, mo->y);
			MSG_WriteLong(&cl->reliablebuf, mo->z);
			MSG_WriteShort(&cl->reliablebuf, mo->type);
			MSG_WriteShort(&cl->reliablebuf, mo->netid);

			// a monster
			if (mo->type == MT_SKULL || (mo->flags & MF_COUNTKILL) )
			{
				MSG_WriteByte(&cl->reliablebuf, svc_mobjangle);
				MSG_WriteShort(&cl->reliablebuf, mo->netid);
				MSG_WriteLong(&cl->reliablebuf, mo->angle);
			}
		}

		// corpses
		if ( mo->health <= 0 && (mo->type == MT_PLAYER 
			|| (mo->flags & MF_COUNTKILL) ) )
		{
			MSG_WriteByte(&cl->reliablebuf, svc_corpse);
			MSG_WriteShort(&cl->reliablebuf, mo->netid);
			MSG_WriteByte (&cl->reliablebuf, mo->frame);
		}

		if (cl->reliablebuf.cursize >= 4096)
		{
			parse_cl = i;
			SV_SendPacket(cl);
		}
    }

	// update sectors
	for (int s=0; s<numsectors; s++)
	{ 
		sector_t* sec = &sectors[s];

	    if (sec->moveable)
		{
			MSG_WriteByte (&cl->reliablebuf, svc_sector);
			MSG_WriteShort (&cl->reliablebuf, s);
			MSG_WriteShort (&cl->reliablebuf, sec->floorheight>>FRACBITS);
			MSG_WriteShort (&cl->reliablebuf, sec->ceilingheight>>FRACBITS);
		}

		if (cl->reliablebuf.cursize >= 4096)
		{
			SV_SendPacket(cl);
		}

	}

	SV_SendPacket(cl);
}

//
// SV_LoadLevel
//
void SV_LoadLevel()
{
	int	i;

	for (i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		// send dmflags
		MSG_WriteByte(&cl->reliablebuf, svc_dmflags);
		MSG_WriteLong(&cl->reliablebuf, dmflags);

		// send skill
		MSG_WriteByte(&cl->reliablebuf, svc_gameskill);
		MSG_WriteByte(&cl->reliablebuf, (byte)gameskill.value);

		player_t *p = (player_t *)&players[i];

		int team = -1;

		// Send special settings like teamplay and deathmatch.
		MSG_WriteByte(&cl->reliablebuf, svc_specialsettings);
		MSG_WriteByte(&cl->reliablebuf, (byte)teamplay.value);
		MSG_WriteByte(&cl->reliablebuf, team);
		MSG_WriteByte(&cl->reliablebuf, (byte)deathmatch.value);

		if(teamplay.value)
		{
			team = SV_ChooseTeamForPlayer(p);
		}
   
		// player must choose red or blue team
		if (teamplay.value && team == -1)
		{
			MSG_WriteByte (&cl->reliablebuf, svc_print);
			MSG_WriteByte (&cl->reliablebuf, PRINT_HIGH);
			MSG_WriteString (&cl->reliablebuf, 
				"Choose red or blue team.\n"
				"Use \"team <teamname>\" command to change your team.\n");

			SV_DropClient();
			return;
		}

		// Force the client to load a new map.
		MSG_WriteByte(&cl->reliablebuf, svc_loadmap);
		MSG_WriteString(&cl->reliablebuf, level.mapname);		
	}
}


void SV_NewLevelClientUpdate()
{
	AActor *mo;
	int		i,j;

	for (i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;
	
		if (deathmatch.value)
			for (j=0; j<MAXPLAYERS; j++)
			{
				if (clients[j].cls < cs_spawned)
					continue;

				MSG_WriteByte(&cl->reliablebuf, svc_updatefrags);
				MSG_WriteByte(&cl->reliablebuf, j);
				MSG_WriteShort(&cl->reliablebuf, players[j].fragcount);
			}

			// update things
			TThinkerIterator<AActor> iterator;
			while ( (mo = iterator.Next() ) )
			{
				if (!mo->netid)
					continue;

				// a monster
				if (mo->type == MT_SKULL || (mo->flags & MF_COUNTKILL) )
				{
					MSG_WriteByte(&cl->reliablebuf, svc_mobjangle);
					MSG_WriteShort(&cl->reliablebuf, mo->netid);
					MSG_WriteLong(&cl->reliablebuf, mo->angle);
				}
			}

			parse_cl = i;
			SV_SendPacket(cl);
	}
}

//
// SV_CalcTeamFrags
//

//
// SV_ChooseTeamForPlayer
//
int SV_ChooseTeamForPlayer(player_t *player)
{
	int index, redcount = 0, bluecount = 0, this_player = -1, fairteam = -1;
	player_t *temp_player;

	if(!teamplay.value)
		return -1;

	if (   !strcmp(player->userinfo.team, "red") 
		|| !strcmp(player->userinfo.team, "RED") 
		|| !strcmp(player->userinfo.team, "Red") ) // Red team
	{
		fairteam = 0;
	}

	if (   !strcmp(player->userinfo.team, "blue") 
		|| !strcmp(player->userinfo.team, "BLUE") 
		|| !strcmp(player->userinfo.team, "Blue") ) // Blue team
	{
		fairteam = 1;
	}

	// DL - if no team provided, stick them in the "fair" team
	for(index = 0; index < MAXPLAYERS; index++)
	{
		if (!playeringame[index])
			continue;

		temp_player = (player_t *)&players[index];

		if(temp_player == player)
		{
			this_player = index;
			continue;
		}

		if (!strcmp(temp_player->userinfo.team, "red") 
			|| !strcmp(temp_player->userinfo.team, "RED") 
			|| !strcmp(temp_player->userinfo.team, "Red") ) // Red team
		{
			redcount ++;
		}

		if (   !strcmp(temp_player->userinfo.team, "blue") 
			|| !strcmp(temp_player->userinfo.team, "BLUE") 
			|| !strcmp(temp_player->userinfo.team, "Blue") ) // Blue team
		{
			bluecount ++;
		}
	}

	if(fairteam == -1)
	{
		fairteam = (redcount <= bluecount) ? 0 : 1;
	}

	if(!fairteam)
	{
		// Red
		strcpy(player->userinfo.team, "RED");
	}
	else
	{
		// Blue
		strcpy(player->userinfo.team, "BLUE");
	}

	if (fairteam == 0)
		player->userinfo.color = V_GetColorFromString (NULL, "ff 00 00");
	else if (fairteam == 1)
		player->userinfo.color = V_GetColorFromString (NULL, "00 00 ff");

	if(this_player != -1)
	{
		// send to other players
		for (int j=0; j<MAXPLAYERS; j++)
		{
			client_t *cl = &clients[j];

			if (cl->cls < cs_spawned)
				continue;

			SV_SendUserInfo(this_player, cl);
		}
	}

	return fairteam;
}


void SV_GetChallenge(void)
{
	int challenge = MSG_ReadFirstLong();

	// DL - todo Spoofing hole temp fix, server could be forced to connect to itself from itself!?
	// Both in and out packets are headed with CHALLENGE, not a good idea!
	//if(net_from.ip[0] == '127' && net_from.ip[1] == '0' && net_from.ip[2] == '0' && net_from.ip[3] == '1')
	//{// todo - fix properly and for multihomed, this is only a placeholder type thing
	//	if(ntohs(net_from.port) == localport)
	//	{
	//		return;
	//	}
	//}

	if (challenge == LAUNCHER_CHALLENGE)
	{
		SV_SendServerInfo();  // for csDoom Launcher
		return;
	}

	if (challenge != CHALLENGE)
		return;

	int n = SV_GetFreeClient();
	if (n==-1)  // a server is full
	{
		SZ_Clear(&tempclient.reliablebuf);
		MSG_WriteByte(&tempclient.reliablebuf, S2C_FULL);
		NET_SendPacket(tempclient.reliablebuf.cursize, tempclient.reliablebuf.data, net_from);
		return;
	}

	client_t *cl = &clients[n];
	cl->address = net_from;
	cl->last_received = gametic;
	cl->cls = cs_challenge;

	SZ_Clear(&cl->reliablebuf);
	MSG_WriteByte(&cl->reliablebuf, S2C_CHALLENGE);
	NET_SendPacket(cl->reliablebuf.cursize, cl->reliablebuf.data, cl->address);
}


void SV_ConnectClient(void)
{
	client_t  *cl;

	// Stray challenge packets?
	if(MSG_ReadFirstLong() == CHALLENGE || MSG_ReadFirstLong() == LAUNCHER_CHALLENGE)
	{
		MSG_ReadLong();
		Printf(PRINT_HIGH, "Ignored duplicate challenge from %s\n", NET_AdrToString(net_from));
		return;
	}

	cl = &clients[parse_cl];
	players[parse_cl].userinfo.netname[0] = 0;
	SZ_Clear(&cl->reliablebuf);

	short version = MSG_ReadShort();
	if (version != VERSION)
	{
		MSG_WriteByte(&cl->reliablebuf, svc_disconnect);
		NET_SendPacket(cl->reliablebuf.cursize, cl->reliablebuf.data, cl->address);
		Printf(PRINT_HIGH, "%s : could not connect, version mismatch (%d)\n", NET_AdrToString(net_from), version);
		SV_DropClient();
		return;
	}

	MSG_WriteByte(&cl->reliablebuf, S2C_CONNECTION);
	NET_SendPacket(cl->reliablebuf.cursize, cl->reliablebuf.data, cl->address);
	cl->cls = cs_connected;

	Printf(PRINT_HIGH, "%s : connect\n", NET_AdrToString(net_from) );
   
	cl->reliable_bps = 0;
	cl->unreliable_bps = 0;
	cl->lastcmdtic = 0;
	cl->allow_rcon = false;

	SZ_Clear(&cl->netbuf);
	SZ_Clear(&cl->reliablebuf);
	SZ_Clear(&cl->relpackets);

	memset(cl->packetseq, -1, sizeof(cl->packetseq) );
	memset(cl->packetbegin, 0, sizeof(cl->packetbegin) );
	memset(cl->packetsize, 0, sizeof(cl->packetsize) );
	cl->sequence = 0;
	cl->last_sequence = -1;
	cl->packetnum = 0;
}


void SV_SpawnClient(void)
{
	client_t *cl = &clients[parse_cl];
	SZ_Clear(&cl->reliablebuf);

	cl->cls = cs_spawned;
	playeringame[parse_cl] = true;

	// send welcome message
	MSG_WriteByte (&cl->reliablebuf, svc_print);
	MSG_WriteByte (&cl->reliablebuf, PRINT_HIGH);
  
	// [NightFang] - Tell the player which type of server this is
	if (teamplay.value)
		MSG_WriteString (&cl->reliablebuf, welcomestring_teamplay);
	else
		MSG_WriteString (&cl->reliablebuf, welcomestring);

	// send server GUID, old clients will ignore it and new clients will know what it means
	MSG_WriteByte (&cl->reliablebuf, svc_print);
	MSG_WriteByte (&cl->reliablebuf, PRINT_HIGH);
	MSG_WriteString (&cl->reliablebuf, guidstring);

	// send welcome message
	MSG_WriteByte (&cl->reliablebuf, svc_print);
	MSG_WriteByte (&cl->reliablebuf, PRINT_HIGH);
	MSG_WriteString (&cl->reliablebuf, motdstring);

	// send dmflags
	MSG_WriteByte(&cl->reliablebuf, svc_dmflags);
	MSG_WriteLong(&cl->reliablebuf, dmflags);

	// send consoleplayer number
	MSG_WriteByte(&cl->reliablebuf, svc_consoleplayer);
	MSG_WriteByte(&cl->reliablebuf, parse_cl);

	// send skill
	MSG_WriteByte(&cl->reliablebuf, svc_gameskill);
	MSG_WriteByte(&cl->reliablebuf, (byte)gameskill.value);

	players[parse_cl].playerstate = PST_REBORN;
	memset (players[parse_cl].frags,0,sizeof(players[parse_cl].frags)); 
	players[parse_cl].fragcount = 0;

	// get client userinfo
	MSG_ReadByte();  // clc_userinfo
	if(!SV_SetupUserInfo()) // send it to other players
	{						// if it's a team game, player's color
		SV_DropClient();	// will be set to red or blue
		return;
	}

	if(cl->cls != cs_spawned)return; // changed - thrown off by userinfo

	cl->rate = MSG_ReadLong(); // get rate value

	player_t *p = (player_t *)&players[parse_cl];

	int team = -1;

	// Send special settings like teamplay and deathmatch.
	MSG_WriteByte(&cl->reliablebuf, svc_specialsettings);
	MSG_WriteByte(&cl->reliablebuf, (byte)teamplay.value);
	MSG_WriteByte(&cl->reliablebuf, team);
	MSG_WriteByte(&cl->reliablebuf, (byte)deathmatch.value);

	if(teamplay.value)
	{
		team = SV_ChooseTeamForPlayer(p);
	}

	// just send to other players
	for (int j=0; j<MAXPLAYERS; j++)
	{
		client_t *cl = &clients[j];

		if (cl->cls < cs_spawned)
			continue;

		SV_SendUserInfo(parse_cl, cl);
	}

	// player must choose red or blue team
	if (teamplay.value && team == -1)
	{
		MSG_WriteByte (&cl->reliablebuf, svc_print);
		MSG_WriteByte (&cl->reliablebuf, PRINT_HIGH);
		MSG_WriteString (&cl->reliablebuf, 
			"Choose red or blue team.\n"
			"Use \"team <teamname>\" command to change your team.\n");

		SV_DropClient();
		return;
	}

	// send a map name
	MSG_WriteByte(&cl->reliablebuf, svc_loadmap);
	MSG_WriteString(&cl->reliablebuf, level.mapname);

	if (deathmatch.value)
		G_DeathMatchSpawnPlayer(parse_cl);
	else
		P_SpawnPlayer(&playerstarts[parse_cl]);
	 
	SV_ClientFullUpdate(cl);

	if (!teamplay.value)
	{
		SV_BroadcastPrintf (PRINT_HIGH, "%s entered the game\n", players[parse_cl].userinfo.netname);
	}
	else
	{ 
		if (team == 0)
			SV_BroadcastPrintf (PRINT_HIGH, "%s entered the game and joined the RED team.\n", players[parse_cl].userinfo.netname);
		if (team == 1)
			SV_BroadcastPrintf (PRINT_HIGH, "%s entered the game and joined the BLUE team.\n", players[parse_cl].userinfo.netname);
	}

	SV_UpdateFrags (&players[parse_cl]);
}


//
// SV_DisconnectClient
//
void SV_DisconnectClient(void)
{
	client_t *cl = &clients[parse_cl];

	// DL - already disconnected
	if(cl->cls == cs_free)
		return;

	playeringame[parse_cl] = false;
	cl->cls = cs_free; // changed - moved
	memset(&cl->address, 0, sizeof(cl->address));
	SZ_Clear(&cl->netbuf);
	SZ_Clear(&cl->reliablebuf);
	SZ_Clear(&cl->relpackets);
    
	if (gametic - cl->last_received == CLIENT_TIMEOUT*35)
		SV_BroadcastPrintf (PRINT_HIGH, "%s timed out\n", players[parse_cl].userinfo.netname);
	else
        SV_BroadcastPrintf (PRINT_HIGH, "client %s disconnected\n", players[parse_cl].userinfo.netname);

	// tell others clients about it		
	for (int i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		MSG_WriteByte(&cl->reliablebuf, svc_disconnectclient);
		MSG_WriteByte(&cl->reliablebuf, parse_cl);
	}

	if(parse_cl < 0)return;
	if(players[parse_cl].mo)players[parse_cl].mo->Destroy();
	players[parse_cl].mo = NULL;

	// DL - and wipe their footsteps off the floor
	memset(&players[parse_cl], 0, sizeof(players[parse_cl]));
}


//
// SV_DropClient
// Called when the player is leaving the server unwillingly. 
//
void SV_DropClient(void)
{
	client_t *cl = &clients[parse_cl];

	MSG_WriteByte(&cl->reliablebuf, svc_disconnect);
	SV_SendPacket(cl);

	SV_DisconnectClient();
}

//
// SV_SendReconnectSignal
// All clients will reconnect. Called when the server changes a map
//
void SV_SendReconnectSignal()
{
	client_t *cl;

	for (int i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];
		
		if (cl->cls < cs_spawned)
			continue;

		MSG_WriteByte(&cl->reliablebuf, svc_reconnect);
		for(int ii = 0; ii < Args.NumArgs(); ii++)
			{
			MSG_WriteString(&cl->reliablebuf, Args.GetArg(ii));
			}
		
		parse_cl = i;
		SV_SendPacket(cl);
	   
		SV_DropClient();	   
		if(players[i].mo)players[i].mo->Destroy();
		players[i].mo = NULL;
	}
parse_cl = -1;
}


void SV_SendDisconnectSignal()
{
	client_t *cl;

	for (int i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];
		
		if (cl->cls < cs_spawned)
			continue;

		parse_cl = i;
		SV_DropClient();

		if(!players[i].mo)
			continue;
		
		players[i].mo->Destroy();
		players[i].mo = NULL;
	}
parse_cl = -1;
}

//
// SV_ExitLevel
// Called when timelimit or fraglimit hit.
//
void SV_ExitLevel()
{
	for (int i=0; i<MAXPLAYERS; i++)
	{
	   client_t *cl = &clients[i];

	   if (cl->cls < cs_spawned)
		   continue;

	   MSG_WriteByte(&cl->reliablebuf, svc_exitlevel);
	}
}

//
// SV_BroadcastPrintf
// Sends text to all active clients.
//
void STACK_ARGS SV_BroadcastPrintf (int level, const char *fmt, ...)
{
    va_list       argptr;
    char          string[4096]; // thanks to Luigi Auriemma for finding an overflow bug here
    client_t     *cl;

    va_start (argptr,fmt);
    vsprintf (string, fmt,argptr);
    va_end (argptr);

    Printf (level, "%s", string);  // print to the console

    for (int i=0; i<MAXPLAYERS; i++)
    {
		cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;        

		MSG_WriteByte (&cl->reliablebuf, svc_print);
		MSG_WriteByte (&cl->reliablebuf, level);
		MSG_WriteString (&cl->reliablebuf, string);
    }
}

//
// SV_BroadcastPrintf
// Sends text to all active clients BUT one.
//
void STACK_ARGS SV_BroadcastPrintfExcluding (int level, int client_nosend, const char *fmt, ...)
{
    va_list       argptr;
    char          string[4096]; // thanks to Luigi Auriemma for finding an overflow bug here
    client_t     *cl;

    va_start (argptr,fmt);
    vsprintf (string, fmt,argptr);
    va_end (argptr);

    Printf (level, "%s", string);  // print to the console

    for (int i=0; i<MAXPLAYERS; i++)
    {
		if(i == client_nosend)
			continue;

		cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;        

		MSG_WriteByte (&cl->reliablebuf, svc_print);
		MSG_WriteByte (&cl->reliablebuf, level);
		MSG_WriteString (&cl->reliablebuf, string);
    }
}

/// @todo clean up
bool bVoting = false;
char cVotemap[50];
bool bVotes[MAXPLAYERS];
unsigned long VoteTimer;

CVAR (sv_votemap, "0", CVAR_ARCHIVE)
CVAR (sv_votemin, "0", CVAR_ARCHIVE)

#include "w_wad.h"

void SV_Votemap(char *cNewMap)
{
	if(!sv_votemap.value)
		return;

	if(bVoting)
		return;

	if(!cNewMap || !strlen(cNewMap) || strlen(cNewMap) > 8)
	{
		MSG_WriteByte(&clients[parse_cl].reliablebuf, svc_print);
		MSG_WriteByte(&clients[parse_cl].reliablebuf, PRINT_HIGH);
		MSG_WriteString(&clients[parse_cl].reliablebuf, "[server]: bad map");
		return;
	}

	if (W_CheckNumForName (cNewMap) == -1)
	{
		MSG_WriteByte(&clients[parse_cl].reliablebuf, svc_print);
		MSG_WriteByte(&clients[parse_cl].reliablebuf, PRINT_HIGH);
		MSG_WriteString(&clients[parse_cl].reliablebuf, "[server]: no such map");
		return;
	}

	SV_BroadcastPrintf (PRINT_CHAT, "[server]: %s called a vote!\n", players[parse_cl].userinfo.netname);
	SV_BroadcastPrintfExcluding (PRINT_CHAT, parse_cl, "[server]: say v to vote for %s\n", cNewMap);

	VoteTimer = level.time;
	bVoting = true;
	memset(&bVotes, 0, sizeof(bVotes));
	bVotes[parse_cl] = true;
	sprintf(cVotemap, "map %s", cNewMap);
}

bool SV_ChangeTeam(char *cTeam)
{
	client_t *cl = &clients[parse_cl];
	player_t *p = (player_t *)&players[parse_cl];

	if(!teamplay.value)
		return false;

	if(level.time - p->userinfo.LastTeamChangeTime < 20*TICRATE)
	{
		MSG_WriteByte (&cl->reliablebuf, svc_print);
		MSG_WriteByte (&cl->reliablebuf, PRINT_HIGH);
		MSG_WriteString (&cl->reliablebuf, "[server]: team change too fast\n");
		return false;	
	}

	if(strncmp(p->userinfo.team, cTeam, strlen(p->userinfo.team) + 1) == NULL)
	{
		MSG_WriteByte (&cl->reliablebuf, svc_print);
		MSG_WriteByte (&cl->reliablebuf, PRINT_CHAT);
		MSG_WriteString (&cl->reliablebuf, "[server]: already on that team\n");
		return false; // Already on that team
	}

		memcpy(p->userinfo.team, cTeam, 7);
		p->userinfo.team[7] = 0;

		int team = -1;
		if(teamplay.value)
		{
			team = SV_ChooseTeamForPlayer(p);
		}

	p->userinfo.LastTeamChangeTime = level.time;

	return true;
}

void SV_Castvote()
{
	if(!bVoting)
		return;

	bVotes[parse_cl] = true;
}

void SV_DoVoting(bool timer)
{
	if(level.time - VoteTimer < 15*TICRATE) // Spend at least a few seconds voting
		return;

	 if(bVoting)
	 {
		 if(!timer || (level.time - VoteTimer > 25*TICRATE))
		 {
			 int iVotes = 0, iPlayers = 0, iNeeded = 0;

			 for(int i = 0; i < MAXPLAYERS; i++)
			 {
				 if(bVotes[i])iVotes++;
				 if(playeringame[i])iPlayers++;
			 }

			 iNeeded = (iPlayers / 2); // 1/2 of players must vote
			 
			 if(iNeeded < sv_votemin.value)
				 iNeeded = sv_votemin.value;

			 if(iVotes >= iNeeded)
			 {
				bVoting = false;
				SV_BroadcastPrintf(PRINT_CHAT, "[server]: vote succeeded\n");
				printf("k %s", cVotemap);AddCommandString(cVotemap);
			 }
			 else
			 {
				 if(timer)
				 {
					bVoting = false;
					SV_BroadcastPrintf(PRINT_CHAT, "[server]: vote failed: %d voted but at least %d votes needed\n", iVotes, iNeeded);
				 }
			 }
		 }
	 }
}

//
// SV_Say
// Show a chat string and send it to others clients.
//
void SV_Say(void)
{
	char     *string;

    string = MSG_ReadString();
	int iLen = strlen(string);

    client_t     *cl;
	cl = &clients[parse_cl];

	if(iLen >= 1000)
		return;

	// DL - New client detected?
	if(strncmp(string, guidstring, strlen(guidstring)) == 0)
	{
		// Will now be treated as a newer client
		clients[parse_cl].ext_version = 1;
		return;
	}

	char votemap[] = "votemap ";
	char mapvote[] = "mapvote ";
	char map[] = "map ";
	char vote[] = "vote ";

	char team[] = "team ";
	char emote[] = "/me ";

	// Emote?
	if(strncmp(string, emote, strlen(emote)) == NULL)
	{
		// Old clients will crash unless the string is specially formatted
		for(int i = 0; i < strlen(string); i++)
			if(string[i] == '%')string[i] = '*';

		SV_BroadcastPrintf (PRINT_CHAT, "* %s %s\n", players[parse_cl].userinfo.netname, string + strlen(emote));
		return;
	}

	// Team change
	if(strncmp(string, team, strlen(team)) == NULL)
	{
		if(SV_ChangeTeam(string + strlen(team)))
			SV_BroadcastPrintf (PRINT_CHAT, "[server]: %s changed team to %s\n", players[parse_cl].userinfo.netname, players[parse_cl].userinfo.team);
		return;
	}

	// A vote called?
	if(strncmp(string, votemap, strlen(votemap)) == NULL)
	{
		SV_Votemap(string + strlen(votemap));
		SV_DoVoting(false);
		return;
	}
	else if(strncmp(string, map, strlen(map)) == NULL)
	{
		SV_Votemap(string + strlen(map));
		SV_DoVoting(false);
		return;
	}
	else if(strncmp(string, vote, strlen(vote)) == NULL)
	{
		SV_Votemap(string + strlen(vote));
		SV_DoVoting(false);
		return;
	}
	else if(strncmp(string, mapvote, strlen(mapvote)) == NULL)
	{
		SV_Votemap(string + strlen(mapvote));
		SV_DoVoting(false);
		return;
	}

	// Vote cast?
	if(bVoting)
		if((string[0] == 'V' || string[0] == 'v') && string[1] == 0)
		{
			SV_Castvote();
			SV_BroadcastPrintf (PRINT_CHAT, "[server]: %s voted\n", players[parse_cl].userinfo.netname);
			SV_DoVoting(false);
			return;
		}

	// Old clients will crash unless the string is specially formatted
	for(int i = 0; i < strlen(string); i++)
		if(string[i] == '%')string[i] = '*';

	SV_BroadcastPrintf (PRINT_CHAT, "%s: %s\n", players[parse_cl].userinfo.netname, string);
}

//
// SV_UpdateMissiles
// Updates missiles position sometimes.
//
void SV_UpdateMissiles(client_t* cl)
{
    AActor *mo;

    TThinkerIterator<AActor> iterator;
    while ( (mo = iterator.Next() ) )
    {
        if (!(mo->flags & MF_MISSILE) )
			continue;

		if (mo->type == MT_PLASMA)
			continue;
            
		// update missile position every 50 tics
		if ( ((gametic + mo->netid) % 50) )
			continue;

		MSG_WriteByte (&cl->netbuf, svc_movemobj);
		MSG_WriteShort (&cl->netbuf, mo->netid);
		MSG_WriteLong (&cl->netbuf, mo->x);
		MSG_WriteLong (&cl->netbuf, mo->y);
		MSG_WriteLong (&cl->netbuf, mo->z);
    }


}


#define MAXCORPSES 15

//
// SV_RemoveCorpses
// Removes some corpses
//
void SV_RemoveCorpses (void)
{
	AActor *mo;
	int     corpses = 0;

	if (gametic%35)
	{
		return;
	}
	else
	{
		TThinkerIterator<AActor> iterator;
		while ( (mo = iterator.Next() ) )
		{
			if (mo->type == MT_PLAYER && (!mo->player || mo->health <=0) )
				corpses++;
		}
	}

	if (corpses > MAXCORPSES)
	{
		TThinkerIterator<AActor> iterator;
		while (corpses > MAXCORPSES && (mo = iterator.Next() ) )
		{
			if (mo->type == MT_PLAYER && !mo->player)
			{
				mo->Destroy();
				corpses--;
			}
		}
	}
}


//
// SV_SendGametic
// Sends gametic to calculate ping
//
void SV_SendGametic(client_t* cl)
{
	if ((gametic%100) != 0)
		return;

	MSG_WriteByte (&cl->reliablebuf, svc_svgametic);
	MSG_WriteLong (&cl->reliablebuf, gametic);
}

// calculates ping using gametic which was sent by SV_SendGametic and
// current gametic
void SV_CalcPing(void)
{	
	int oldtic = MSG_ReadLong();

	int ping = gametic - oldtic;
	if (ping > 255) ping = 255;
	players[parse_cl].ping = ping;
}

//
// SV_UpdatePing
// send pings to a client
//
void SV_UpdatePing(client_t* cl)
{
	if ((gametic%101) != 0)
		return;

	if (cl->cls < cs_spawned)
		return;

	for (unsigned char j=0; j<MAXPLAYERS; j++)
	{
		if (clients[j].cls < cs_spawned)
			continue;

	    MSG_WriteByte(&cl->reliablebuf, svc_updateping);
	    MSG_WriteByte(&cl->reliablebuf, j);  // player
	    MSG_WriteByte(&cl->reliablebuf, players[j].ping);
	}
}


//
// SV_UpdateDeadPlayers
// Update player's frame while he's dying.
//
void SV_UpdateDeadPlayers()
{
    AActor *mo;

    TThinkerIterator<AActor> iterator;
    while ( (mo = iterator.Next() ) )
    {
        if (mo->type != MT_PLAYER || mo->player)
			continue;
		
		if (mo->oldframe != mo->frame)
			for (int i=0; i<MAXPLAYERS; i++)
			{
				client_t *cl = &clients[i];

				if (cl->cls < cs_spawned)
					continue;

				MSG_WriteByte (&cl->reliablebuf, svc_mobjframe);
				MSG_WriteShort (&cl->reliablebuf, mo->netid);
				MSG_WriteByte (&cl->reliablebuf, mo->frame);
			}
		
		mo->oldframe = mo->frame;
    }

}


//
// SV_UpdateSectors
// Update doors, moving floors, ceilings etc...
//
void SV_UpdateSectors(client_t* cl)
{
	for (int s=0; s<numsectors; s++)
	{ 
		sector_t* sec = &sectors[s];

		if (sec->floordata && sec->floordata->IsKindOf (RUNTIME_CLASS(DMover)))
		{
			MSG_WriteByte (&cl->netbuf, svc_floorheight);
			MSG_WriteShort (&cl->netbuf, s);
			MSG_WriteShort (&cl->netbuf, sec->floorheight>>FRACBITS);
		}
		
		if (sec->ceilingdata && sec->ceilingdata->IsKindOf (RUNTIME_CLASS(DMover)))
		{
			MSG_WriteByte (&cl->netbuf, svc_ceilingheight);
			MSG_WriteShort (&cl->netbuf, s);
			MSG_WriteShort (&cl->netbuf, sec->ceilingheight>>FRACBITS);
		}
	}
}

//
// SV_ClearClientsBPS
//
void SV_ClearClientsBPS(void)
{
	if (gametic % 35)
		return;

	for (int i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		cl->reliable_bps = 0;
		cl->unreliable_bps = 0;
	}
}


//
// SV_WriteCommands
//
void SV_WriteCommands(void)
{
	int        i, j;
	client_t  *cl;

	
	SV_UpdateDeadPlayers(); // Update dying players.

	for (i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;
		
		// Don't need to update origin every tic.
		// The server sends origin and velocity of a 
		// player and the client always knows origin
		// on the next tic.
		if (gametic % 2)
			for (j=0; j<MAXPLAYERS; j++)
			{
				if (!playeringame[j] || j == i)
					continue;

				AActor *mobj = players[j].mo;
				if(!mobj)
					continue;

				byte bits = 0;

				MSG_WriteByte(&cl->netbuf, svc_moveplayer);
				MSG_WriteByte(&cl->netbuf, j);     // player number
				
				if (mobj->x != mobj->oldx)
					bits |= CM_X;
				if (mobj->y != mobj->oldy)
					bits |= CM_Y;
				if (mobj->z != mobj->oldz)
					bits |= CM_Z;
				if (mobj->angle != mobj->oldangle)
					bits |= CM_ANGLE;
				if (mobj->frame != mobj->oldframe)
					bits |= CM_FRAME;
				if (mobj->momx != mobj->oldmomx)
					bits |= CM_MOMX;
				if (mobj->momy != mobj->oldmomy)
					bits |= CM_MOMY;
				if (mobj->momz != mobj->oldmomz)
					bits |= CM_MOMZ;

				MSG_WriteByte(&cl->netbuf, bits);

				if (mobj->x != mobj->oldx)
					MSG_WriteLong(&cl->netbuf, mobj->x);				
				if (mobj->y != mobj->oldy)
					MSG_WriteLong(&cl->netbuf, mobj->y);
				if (mobj->z != mobj->oldz)
					MSG_WriteLong(&cl->netbuf, mobj->z);
				
				if (mobj->angle != mobj->oldangle)
					MSG_WriteLong(&cl->netbuf, mobj->angle);

				if (mobj->frame != mobj->oldframe)
				{
					if (mobj->frame == 32773)
						MSG_WriteByte(&cl->netbuf, PLAYER_FULLBRIGHTFRAME);
					else
						MSG_WriteByte(&cl->netbuf, mobj->frame);
				}

				// write velocity
				if (mobj->momx != mobj->oldmomx)
					MSG_WriteLong(&cl->netbuf, mobj->momx);
				if (mobj->momy != mobj->oldmomy)
					MSG_WriteLong(&cl->netbuf, mobj->momy);
				if (mobj->momz != mobj->oldmomz)
					MSG_WriteLong(&cl->netbuf, mobj->momz);
			}


		SV_UpdateConsolePlayer(cl);

		SV_UpdateMissiles(cl);

        SV_UpdateSectors(cl); // update doors, moving floors, ceilings etc		                        

	 	SV_SendGametic(cl); // used to calculate a ping value,
		                    // when a client returns it
     	SV_UpdatePing(cl);
	}

	// save players' origin, velocity, frame and angle
	if (gametic % 2)
		for (i=0; i<MAXPLAYERS; i++)
		{
			if (!playeringame[i])
				continue;

			AActor *mobj = players[i].mo;

			if(!mobj)
				continue;
		
			mobj->oldx = mobj->x;
			mobj->oldy = mobj->y;
			mobj->oldz = mobj->z;
			mobj->oldangle = mobj->angle;
			mobj->oldframe = mobj->frame;
			mobj->oldmomx = mobj->momx;
			mobj->oldmomy = mobj->momy;
			mobj->oldmomz = mobj->momz;
		}
}

void SV_PlayerTriedToCheat()
{
	SV_BroadcastPrintf(PRINT_HIGH, "%s tried to cheat!\n", players[parse_cl].userinfo.netname);
	SV_DropClient();
}

/*
struct usercmd_s {
	byte	msec;			// not sure how to use this yet...
	byte	buttons;
	short	pitch;			// up/down. currently just a y-sheering amount
	short	yaw;			// left/right	// If you haven't guessed, I just
	short	roll;			// tilt			// ripped these from Quake2's usercmd.
	short	forwardmove;
	short	sidemove;
	short	upmove;
	byte	impulse;
	byte	use;
};
*/

void SV_GetPlayerCmd(void)
{
	client_t *cl = &clients[parse_cl];
	ticcmd_t *cmd = &players[parse_cl].cmd;
	int       tic;

    tic = MSG_ReadLong(); // get client gametic

	// get the previous cmds and angle
	/*if (  cl->lastcmdtic && (tic - cl->lastcmdtic >= 2) 
		&& gamestate != GS_INTERMISSION )
	{
		cmd->ucmd.buttons = MSG_ReadByte();
	    players[parse_cl].mo->angle = MSG_ReadShort() << 16;
		
		if (dmflags & DF_NO_FREELOOK)
		{
			players[parse_cl].mo->pitch = 0;
			MSG_ReadShort();
		}
		else
			players[parse_cl].mo->pitch = MSG_ReadShort() << 16;

	    cmd->ucmd.forwardmove = MSG_ReadShort();
	    cmd->ucmd.sidemove = MSG_ReadShort();

		if ( abs(cmd->ucmd.forwardmove) > 12800 
			|| abs(cmd->ucmd.sidemove) > 12800)
		{
			SV_PlayerTriedToCheat();
			return;
		}

		cmd->ucmd.impulse = MSG_ReadByte();

		P_PlayerThink (&players[parse_cl]);
	    players[parse_cl].mo->RunThink();
	}
	else*/
	{
		// get 10 bytes
		MSG_ReadLong();
		MSG_ReadLong();
		MSG_ReadShort();
	}

	// get current cmds and angle
	cmd->ucmd.buttons = MSG_ReadByte();
	if (gamestate != GS_INTERMISSION)
	{
		players[parse_cl].mo->angle = MSG_ReadShort() << 16;
		
		if (dmflags & DF_NO_FREELOOK)
		{
			players[parse_cl].mo->pitch = 0;
			MSG_ReadShort();
		}
		else
			players[parse_cl].mo->pitch = MSG_ReadShort() << 16;
	}
	else
		MSG_ReadLong();

	cmd->ucmd.forwardmove = MSG_ReadShort();
	cmd->ucmd.sidemove = MSG_ReadShort();
	
	if ( abs(cmd->ucmd.forwardmove) > 12800 
		|| abs(cmd->ucmd.sidemove) > 12800)
	{
		SV_PlayerTriedToCheat();
		return;
	}
	
	cmd->ucmd.impulse = MSG_ReadByte();

	if (gamestate != GS_INTERMISSION)
	{
		P_PlayerThink (&players[parse_cl]);
		
		// DL
		if(players[parse_cl].mo)
			players[parse_cl].mo->RunThink();
	}

	cl->lastcmdtic = tic;
}


void SV_UpdateConsolePlayer(client_t* cl)
{
	int        playernum;
	player_t  *player;

	// It's not a good idea to send 33 bytes every tic.
	if (gametic % 3)
		return;

	for (playernum=0; playernum<MAXPLAYERS; playernum++)
	{
		if (cl->cls < cs_spawned)
			continue;

		if (cl == &clients[playernum])
		{
			player = &players[playernum];
			break;
		}
	}
	
	if (!player)
	    return;
	
	// haven't got any cmds yet
	if (!cl->lastcmdtic)
		return;

	// no mo?
	if(!player->mo)
		return;

	// client player will update his position if packets were missed
	MSG_WriteByte (&cl->netbuf, svc_updatelocalplayer);
	MSG_WriteLong (&cl->netbuf, cl->lastcmdtic);

	MSG_WriteLong (&cl->netbuf, player->mo->x);
	MSG_WriteLong (&cl->netbuf, player->mo->y);
	MSG_WriteLong (&cl->netbuf, player->mo->z);

	MSG_WriteLong (&cl->netbuf, player->mo->momx);
	MSG_WriteLong (&cl->netbuf, player->mo->momy);
	MSG_WriteLong (&cl->netbuf, player->mo->momz);

//	MSG_WriteShort (&cl->netbuf, player->mo->momx >> FRACBITS);
//	MSG_WriteShort (&cl->netbuf, player->mo->momy >> FRACBITS);
//	MSG_WriteShort (&cl->netbuf, player->mo->momz >> FRACBITS);
}


//
// SV_ParseCommands
//
void SV_ParseCommands(void)
{		
	 int  cmd;
	 char *str;
	 
	 while(1)	 
	 {  
 		if (msg_badread)
		{
            Printf (PRINT_HIGH, "SV_ParseCommands: badread\n");
          	SV_DropClient();
            return;
		}

		// if already dropped move on
		if(clients[parse_cl].cls < cs_spawned)break; // changed - added

		cmd = MSG_ReadByte();

		if (cmd == -1) // EOM
			break;

		if (clients[parse_cl].cls < cs_spawned)
			break;

		if(!players[parse_cl].mo)
			break;

		switch(cmd)
		{
		case clc_disconnect: // ok
			SV_DropClient();
			return;

        case clc_userinfo: // risky, but unlikely to crash
			if(!SV_SetupUserInfo())
				return;

			// just send to other players
			for (int j=0; j<MAXPLAYERS; j++)
			{
				client_t *cl = &clients[j];

				if (cl->cls < cs_spawned)
					continue;

				SV_SendUserInfo(parse_cl, cl);
			}

			break;

		case clc_say: // potentially crashes, but unlikely to cause overflow first - fixed
			SV_Say();
			break;

		case clc_move: // ok
			SV_GetPlayerCmd();
			break;

		case clc_svgametic: // ok
			SV_CalcPing();
			break;

		case clc_rate: // ok
			clients[parse_cl].rate = MSG_ReadLong();
			// DL - prevent problems by locking rate within a range
			if(clients[parse_cl].rate < 500)clients[parse_cl].rate = 500;
			if(clients[parse_cl].rate > 8000)clients[parse_cl].rate = 8000;
			break;

		case clc_ack: // potentially crashes
			SV_AcknowledgePacket();
			break;
	
		case clc_rcon: // ok
			str = MSG_ReadString();
			if (clients[parse_cl].allow_rcon)
				AddCommandString (str);
			break;

		case clc_rcon_password: // ok
			str = MSG_ReadString();
			if (strlen(rcon_password.string) >= 1 && !strcmp(rcon_password.string, str) )
			{
				clients[parse_cl].allow_rcon = true;
				Printf (PRINT_HIGH, "%s has logged in with rcon\n", players[parse_cl].userinfo.netname);
			}
			else
			{
				Printf (PRINT_HIGH, "%s: invalid rcon password\n", players[parse_cl].userinfo.netname);
			}
			break;

		case svc_launcher_challenge:
			MSG_ReadByte();
			MSG_ReadByte();
			MSG_ReadByte();
			if(MSG_ReadFirstLong() == LAUNCHER_CHALLENGE)
			{
				Printf(PRINT_HIGH, "SV_ParseCommands: Launcher refresh from connected client\n", cmd);
			}
			break;

		case svc_challenge:
			MSG_ReadByte();
			MSG_ReadByte();
			MSG_ReadByte();
			if(MSG_ReadFirstLong() == CHALLENGE)
			{
				Printf(PRINT_HIGH, "SV_ParseCommands: Challenge from connected client\n", cmd);
				clients[parse_cl].cls = cs_challenge;
				
				MSG_WriteByte(&clients[parse_cl].reliablebuf, S2C_CHALLENGE);
				NET_SendPacket(clients[parse_cl].reliablebuf.cursize, clients[parse_cl].reliablebuf.data, clients[parse_cl].address);
			}
			break;
		
		default: // ok
			Printf(PRINT_HIGH, "SV_ParseCommands: Unknown client message (%d).\n", cmd);
          	SV_DropClient();
            return;
		}	 
	}

	SV_DoVoting(true);
}


//
// SV_RunTics
//
void SV_RunTics (void)
{
	int              nowtime;
	int              newtics;

	I_DoSelect();

	nowtime = I_GetTime ();
        newtics = nowtime - gametime;

	// console input
	char *cmd = I_ConsoleInput();
	if (cmd)
	{
		AddCommandString (cmd);
	}

	SV_GetPackets(); // get packets

	if (newtics > 0)
	{
		DObject::BeginFrame ();

		// run the newtime tics
		while (newtics--)
		{
			C_Ticker ();
			G_Ticker ();
			SV_UpdateMaster();			
			gametic++;
		}

		DObject::EndFrame ();
	
		gametime = nowtime;
	}
}



//////// FRAGS LOGGING ///////////

FILE  *sv_fraglogfile;

//
// SV_Fraglogfile
// from quake world
//
void SV_Fraglogfile (void)
{
	char	name[32];
	int		i;

	if (sv_fraglogfile)
	{
		Printf (PRINT_HIGH, "Frag file logging off.\n");
		fclose (sv_fraglogfile);
		sv_fraglogfile = NULL;
		return;
	}

	// find an unused name
	for (i=0 ; i<1000 ; i++)
	{
		sprintf (name, "frag_%i.log", i);
		sv_fraglogfile = fopen (name, "r");
		if (!sv_fraglogfile)
		{	// can't read it, so create this one
			sv_fraglogfile = fopen (name, "w");
			if (!sv_fraglogfile)
				i=1000;	// give error
			break;
		}
		fclose (sv_fraglogfile);
	}

	if (i==1000)
	{
		Printf (PRINT_HIGH, "Can't open any logfiles.\n");
		sv_fraglogfile = NULL;
		return;
	}

	Printf (PRINT_HIGH, "Logging frags to %s.\n", name);
}


//
// SV_logfrag
//
void SV_Logfrag (player_t* source, player_t* target)
{
	if (!sv_fraglogfile) 
		return;

	fprintf (sv_fraglogfile, "\\%s\\%s\\\n",
		source->userinfo.netname, target->userinfo.netname);
	fflush (sv_fraglogfile);
}